"""
   Rapid identification of taxonomic assignments (RITA)
   This program executes and combines the results of several BLAST algorithms and
   the Naive Bayes algorithm.  Sequences that are classified early in the pipeline
   are not executed with later algorithms, saving runtime.
   
   This work has been submitted for publication:
   MacDonald Norman J, Parks Donovan H, and Beiko Robert G.  Rapid identification
   of taxonomic assignments. Submitted 2011.
   
   If you have any questions or bug reports, please let us know at <beiko@cs.dal.ca>.
   
   This program depends on the installation of the Fragment Classification Package
   (available at http://kiwi.cs.dal.ca/Software_main)

   classifier.py contains interfaces for running each classifier (NB, BLAST algos).  They also try to check
   for file corruption and re-execute the classifier as required. None of these objects label
   fragments, they simply grab unfinished fragments, run the classifier, and output to an appropriate
   log file.  It is the job of the labellers in 'labeller.py' to read the log files and determine
   a classification for each fragment.
"""

from logio import BLASTLogIO, NBLogIO
from rita.core.sequence import SequenceIO
from rita.core.fastaio import FastaIO

import os
from datetime import datetime


class Classifier(object):
    name = "Classifier"

    def write_sequences_to_scratch_directory(self,job):
        scratch_dir = job['scratch_directory']
        ufilename = "job%d_%s.fasta"%(job['job_id'],self.name)
        ufilepath = "%s/%s"%(scratch_dir,ufilename)
        sio = SequenceIO()
        ufilepath = sio.write_unfinished(job,ufilepath)
        self.nseqs = 0
        return ufilepath
    
    def is_finished(self,job):
        finished = job.get('%s_isFinished'%(self.name),False)
        return finished

    def set_finished(self,job):
        job['%s_isFinished'%(self.name)] = True
        
    def set_proxy(self,proxy):
        self.proxy = proxy
    
    def start_timer(self):
        self.start_time = datetime.now()

    def stop_timer(self,job):
        end_time = datetime.now()
        delta = end_time - self.start_time
        seconds = delta.seconds + (delta.days*24*3600)
        job.add_runtime(self.name,seconds,self.nseqs)
        
        
    def run(self,pid,np):
        pass


class NbClassifier(Classifier):
    name = 'NB'
    def run(self,job):
        """ Run, parse results, save somewhere"""
        """ This should be split by models, not sequences """
        
        filepath = self.write_sequences_to_scratch_directory(job)
        output_filename = "%s.%s.%s.results"%(job['sequence_filepath'],job['job_id'],self.name)
            
        curpath = os.getcwd()
        command = """%s -q %s/%s -m %s -r %s/%s -e .%s.txt"""%(job['nb_path'],curpath,filepath,job['nb_model_rel_filepath'],curpath,output_filename,job['job_id'])
        print command
        finished = False
        while not finished:
            finished = True
            self.start_timer()
            curpath = os.getcwd()
            os.chdir(job['fcp_root'])
            os.system(command)
            os.chdir(curpath)
            
            self.stop_timer(job)
        
            logio = NBLogIO('%s/%s'%(curpath,output_filename))
            if logio.is_corrupt():
                finished = False
                print('Corrupt NB log.  Redoing...')

            
    

class DbnClassifier(Classifier):
    name = 'DCMEGABLAST'
    def run(self,job):
        """ Run, parse results, save somewhere"""
        filepath = self.write_sequences_to_scratch_directory(job)
        output_files = []
        full_output_filename = "%s.%s.%s.results"%(job['sequence_filepath'],job['job_id'],self.name)

        for i in range(int(job['blastdb_parts'])):
          output_filename = "%s.%s"%(full_output_filename,i+1)
          output_files.append(output_filename)
          blastdbname = job['nucleotide_blastdb_path']%(i+1)
          command = """%s -task dc-megablast -db %s -query %s -outfmt '6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore' -evalue %s -out %s"""%(job['blastn_path'],blastdbname,filepath,job['dcmegablast_evalue'],output_filename)
          finished = False
          while not finished:
            finished = True

            self.start_timer()
            os.system(command)
            self.stop_timer(job)
            logio = BLASTLogIO(output_filename)
            
            if logio.is_corrupt():
                finished = False
                print('Corrupt BLAST log.  Redoing...')
        #Put seqs back in sorted order
        fout = open(full_output_filename,'w')
        filehandles = []
        for filename in output_files:
           filehandles.append(open(filename))
        lastline = [None]*len(filehandles)
        fio = FastaIO()
        for seqid,seq in fio.read(filepath):
           chandle = 0
           while chandle < len(filehandles):
             try:
                if lastline[chandle] != None:
                  nl = lastline[chandle]
                  lastline[chandle] = None
                else:
                  nl = next(filehandles[chandle])
                sid = nl.split('\t')[0]
                if sid != seqid:
                   lastline[chandle] = nl
                   chandle += 1
                else:
                   fout.write(nl)
             except StopIteration:
                lastline[chandle] = None
                chandle += 1
        for f in filehandles:
           f.close()
        fout.close()
               
         

class BnClassifier(Classifier):
    name = 'BLASTN'
    def run(self,job):
        filepath = self.write_sequences_to_scratch_directory(job)
        output_files = []
        full_output_filename = "%s.%s.%s.results"%(job['sequence_filepath'],job['job_id'],self.name)

        for i in range(int(job['blastdb_parts'])):
          output_filename = "%s.%s"%(full_output_filename,i+1)
          output_files.append(output_filename)
          blastdbname = job['nucleotide_blastdb_path']%(i+1)

          command = """%s -task blastn -db %s -query %s -outfmt '6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore' -evalue %s -out %s"""%(job['blastn_path'],blastdbname,filepath,job['blastn_evalue'],output_filename)
          finished = False
          while not finished:
            finished = True
        
        
            self.start_timer()
            os.system(command)
            self.stop_timer(job)
            logio = BLASTLogIO(output_filename)
            if logio.is_corrupt():
                finished = False
                print('Corrupt BLASTN log.  Redoing...')
        #Put seqs back in sorted order
        fout = open(full_output_filename,'w')
        filehandles = []
        for filename in output_files:
           filehandles.append(open(filename))
        lastline = [None]*len(filehandles)
        fio = FastaIO()
        for seqid,seq in fio.read(filepath):
           chandle = 0
           while chandle < len(filehandles):
             try:
                if lastline[chandle] != None:
                  nl = lastline[chandle]
                  lastline[chandle] = None
                else:
                  nl = next(filehandles[chandle])
                sid = nl.split('\t')[0]
                if sid != seqid:
                   lastline[chandle] = nl
                   chandle += 1
                else:
                   fout.write(nl)
             except StopIteration:
                lastline[chandle] = None
                chandle += 1
        for f in filehandles:
           f.close()
        fout.close()

                
class BxClassifier(Classifier):
    name = 'BLASTX'
    def run(self,job):
        filepath = self.write_sequences_to_scratch_directory(job)
        output_files = []
        full_output_filename = "%s.%s.%s.results"%(job['sequence_filepath'],job['job_id'],self.name)

        for i in range(int(job['blastdb_parts'])):
          output_filename = "%s.%s"%(full_output_filename,i+1)
          output_files.append(output_filename)
          blastdbname = job['protein_blastdb_path']%(i+1)
        
          finished = False
          while not finished:
            finished = True
            command = """%s -db %s -query %s -outfmt '6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore' -evalue %s -out %s"""%(job['blastx_path'],blastdbname,filepath,job['blastx_evalue'],output_filename)
        
            self.start_timer()
            os.system(command)
            self.stop_timer(job)
            logio = BLASTLogIO(output_filename)
            if logio.is_corrupt():
                finished = False
                print('Corrupt BLASTX log.  Redoing...')
        #Put seqs back in sorted order
        fout = open(full_output_filename,'w')
        filehandles = []
        for filename in output_files:
           filehandles.append(open(filename))
        lastline = [None]*len(filehandles)
        fio = FastaIO()
        for seqid,seq in fio.read(filepath):
           chandle = 0
           while chandle < len(filehandles):
             try:
                if lastline[chandle] != None:
                  nl = lastline[chandle]
                  lastline[chandle] = None
                else:
                  nl = next(filehandles[chandle])
                sid = nl.split('\t')[0]
                if sid != seqid:
                   lastline[chandle] = nl
                   chandle += 1
                else:
                   fout.write(nl)
             except StopIteration:
                lastline[chandle] = None
                chandle += 1
        for f in filehandles:
           f.close()
        fout.close()

                
class UBxClassifier(Classifier):
    """This classifier first runs UBLASTX to get a set of matches.
       Since longer matches can easily have an E-value of zero, we build a blastx
       database with only the hits identified by ublastx.
    """
    name = 'UBLASTX'
    def run(self,job):
        filepath = self.write_sequences_to_scratch_directory(job)
        output_files = []
        full_output_filename = "%s.%s.%s.results"%(job['sequence_filepath'],job['job_id'],self.name)
        all_models = set()
        for i in range(int(job['ublastdb_parts'])):
          output_filename = "%s.%s"%(full_output_filename,i+1)
          output_files.append(output_filename)
          blastdbname = job['protein_ublastdb_path']%(i+1)
        
          finished = False
          while not finished:
            finished = True
            command = """%s --quiet --db %s --query %s --evalue %s --blast6out %s"""%(job['ublastx_path'],blastdbname,filepath,job['ublastx_evalue'],output_filename)
        
            self.start_timer()
            os.system(command)
            self.stop_timer(job)
            logio = BLASTLogIO(output_filename)
            for fragid,scores in logio.read():
                all_models.update(set([x[1] for x in scores]))
        #Now go through all of the sequences and extract the ones we want for our temporary ublast db
        outputdbfasta = '%s.%s.%s.ublastxfiltered.fasta'%(job['sequence_filepath'],job['job_id'],self.name)
        fout = open(outputdbfasta,'w')
        fio = FastaIO()
        for i in range(int(job['ublastdb_parts'])):
            blastdbname = job['protein_ublastdb_path']%(i+1)
            for seqid,seq in fio.read(blastdbname):
                if seqid in all_models:
                    fout.write('>%s\n%s\n'%(seqid,seq))
        fout.close()
        #Build the blastx database
        outputdb = '%s.db'%(outputdbfasta)
        command = """%s -in %s -out %s -dbtype prot"""%(job['makeblastdb_path'],outputdbfasta,outputdb)
        os.system(command)
        """Now run all sequences against the new database """
        #output_filename = '%s.blastx'%(full_output_filename)
        blastdbname = outputdb
        
        finished = False
        while not finished:
            finished = True
            command = """%s -db %s -query %s -outfmt '6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore' -evalue %s -out %s"""%(job['blastx_path'],blastdbname,filepath,job['blastx_evalue'],full_output_filename)
        
            self.start_timer()
            os.system(command)
            self.stop_timer(job)
            logio = BLASTLogIO(full_output_filename)
            if logio.is_corrupt():
                finished = False
                print('Corrupt BLASTX log.  Redoing...')
        
        
        

