"""
   Rapid identification of taxonomic assignments (RITA)
   This program executes and combines the results of several BLAST algorithms and
   the Naive Bayes algorithm.  Sequences that are classified early in the pipeline
   are not executed with later algorithms, saving runtime.
   
   This work has been submitted for publication:
   MacDonald Norman J, Parks Donovan H, and Beiko Robert G.  Rapid identification
   of taxonomic assignments. Submitted 2011.
   
   If you have any questions or bug reports, please let us know at <beiko@cs.dal.ca>.
   
   This program depends on the installation of the Fragment Classification Package
   (available at http://kiwi.cs.dal.ca/Software_main)

   labeller.py contains objects that read appropriate classifier logs and detect if
   the criteria for labelling each sequence has been met.  If it has, the fragment
   is labelled and is set to 'finished' and does not move on to the next labeller in the pipeline
   
   Labellers are the fundamental building block of the pipeline.  Each labeller must have
   a static prerequisites variable set to the classifier objects that are required to be run before it
   can label each fragment.
"""

from classifier import *
from rita.core.sequence import SequenceIO
from logio import NBLogIO, BLASTLogIO
from rita.core.taxonomy import TabTaxonomy
import math

class Labeller(object):
    # A list of Classifiers that are prerequisites to labelling
    prerequisites = []

    def label(self,job,group_id):
        pass

    def finish(self,job,fragment_id,score_model,group):
        job.write_classification(fragment_id,score_model[1],score_model[0],group,score_model[2])

class NullLabeller(Labeller):
    def label(self,job,group_id):
        'Output remaining sequences as None'
        seqio = SequenceIO()
        unfinished = seqio.get_unfinished(job)        
        for fragid in unfinished:
            self.finish(job,fragid,['None','None','None'],group_id)

class MLLabeller(Labeller):
    def label(self,job,group_id):
        logfilename = '%s.%s.%s.results'%(job['sequence_filepath'],job['job_id'],self.homology_algorithm)
        mask = None
        maskloo = {}
        if job['maskfile'] != None:
           maskloo = {}
           for line in open(job['maskfile']):
              fields = line.strip().split('\t')
              maskloo[fields[0]] = set(fields[1:])
        if job['speciesmask'] != None:
           mask = TabTaxonomy(job['speciesmask'])
           mask.update(job['taxonomy'])

        self.fio = self.fiot(logfilename,mask,maskloo=maskloo)
        # Go through NB and the BLAST algorithm files
        # Label seqs that are the best - set htem to finished in job
        
        rank = job['rank']
        it = self.fio.read()
        finished = job.get_finished()
        tax = job['taxonomy']
        jr = job['rank']
        try:
            while True:
                record = next(it)
                while record != None and record[0] in finished:
                    record = next(it)
                if record == None:
                    break
                if jr == 'FLEXIBLE':
                    #job['rank'] = 'SPECIES'
                    #blbest = self.fio.get_best(job,record)
                    if len(record[1]) > 0:
                      rank = job['proxy'].get(tax.get(record[1][0][1],'SPECIES'),'DOMAIN')
                      job['rank'] = rank
                
                bnb = self.fio.get_best_next_best(job,record)
                if bnb[0] != None:
                    self.finish(job,record[0],bnb[0],group_id)
                job['rank'] = jr
        except StopIteration:
            pass


class NbMLLabeller(MLLabeller):
    prerequisites = [NbClassifier]
    homology_algorithm = 'NB'
    def __init__(self):
        self.fiot = NBLogIO
    
class NbHomologyLabeller(Labeller):
    def label(self,job,group_id):
        # Go through NB and the BLAST algorithm files
        # Label seqs that are the best - set htem to finished in job
        nblogfilename = '%s.%s.%s.results'%(job['sequence_filepath'],job['job_id'],'NB')
        blastlogfilename = '%s.%s.%s.results'%(job['sequence_filepath'],job['job_id'],self.homology_algorithm)
        mask = None
        maskloo = {}
        if job['speciesmask'] != None:
           mask = TabTaxonomy(job['speciesmask'])
           mask.update(job['taxonomy'])
        if job['maskfile'] != None:
           maskloo = {}
           for line in open(job['maskfile']):
              fields = line.strip().split('\t')
              maskloo[fields[0]] = set(fields[1:])

        nbio = NBLogIO(nblogfilename,mask,maskloo=maskloo)
        bio = BLASTLogIO(blastlogfilename,mask,maskloo=maskloo)
        itblast = bio.read()
        itnb = nbio.read()
        finished = job.get_finished()
        jr = job['rank']
        tax = job['taxonomy']
        try:
            while True:
                blastrecord = next(itblast)
                while blastrecord != None and blastrecord[0] in finished:
                    blastrecord = next(itblast)
                if blastrecord == None:
                    break
                nbrecord = next(itnb)
                while nbrecord != None and blastrecord[0] != nbrecord[0]:
                    nbrecord =next(itnb)
                if jr == 'FLEXIBLE':
                    #job['rank'] = 'SPECIES'
                    #blbest = bio.get_best(job,blastrecord)
                    if len(blastrecord[1]) > 0:
                      rank = job['proxy'].get(tax.get(blastrecord[1][0][1],'SPECIES'),'DOMAIN')
                      job['rank'] = rank
                blbest = bio.get_best(job,blastrecord)
                nbbest = nbio.get_best(job,nbrecord)
                
                if nbbest != None and blbest != None and nbbest[1] == blbest[1]:
                    self.finish(job,nbrecord[0],nbbest,group_id)
                #elif nbbest != None and blbest != None:
                #    print 'rank %s %s != %s'%(job['rank'],nbbest[1],blbest[1])
                job['rank'] = jr
        except StopIteration:
            pass
        job['rank'] = jr
        
        

class NbDbnLabeller(NbHomologyLabeller):
    prerequisites = [NbClassifier,DbnClassifier]
    homology_algorithm = 'DCMEGABLAST'

class NbBnLabeller(NbHomologyLabeller):
    prerequisites = [NbClassifier,BnClassifier]
    homology_algorithm = 'BLASTN'
        
class NbBxLabeller(NbHomologyLabeller):
    prerequisites = [NbClassifier,BxClassifier]
    homology_algorithm = 'BLASTX'
    

class NbUBxLabeller(NbHomologyLabeller):
    prerequisites = [NbClassifier,UBxClassifier]
    homology_algorithm = 'UBLASTX'
    


class HomologyRatioLabeller(Labeller):
    def label(self,job,group_id):
        logfilename = '%s.%s.%s.results'%(job['sequence_filepath'],job['job_id'],self.homology_algorithm)
        mask = None
        maskloo = {}
        if job['maskfile'] != None:
           maskloo = {}
           for line in open(job['maskfile']):
              fields = line.strip().split('\t')
              maskloo[fields[0]] = set(fields[1:])
        if job['speciesmask'] != None:
           mask = TabTaxonomy(job['speciesmask'])
           mask.update(job['taxonomy'])

        self.fio = self.fiot(logfilename,mask,maskloo=maskloo)
        # Go through NB and the BLAST algorithm files
        # Label seqs that are the best - set htem to finished in job
        diff = job['%s_diff'%(self.__class__.homology_algorithm.lower())]
        rank = job['rank']
        it = self.fio.read()
        finished = job.get_finished()
        tax = job['taxonomy']
        jr  = job['rank']
        try:
            while True:
                record = next(it)
                while record != None and record[0] in finished:
                    record = next(it)
                if record == None:
                    break
                if jr == 'FLEXIBLE':
                    #job['rank'] = 'SPECIES'
                    #blbest = self.fio.get_best(job,record)
                    if len(record[1]) > 0:
                      rank = job['proxy'].get(tax.get(record[1][0][1],'SPECIES'),'DOMAIN')
                      job['rank'] = rank
                bnb = self.fio.get_best_next_best(job,record)
                if bnb[0] != None:
                    if bnb[1] == None:
                        #if self.homology_algorithm == 'DCMEGABLAST':
                        #        print '%s\t%s\t%s\t%s'%(record[0],bnb[0][1],bnb[0][2],bnb[0][0])
                        self.finish(job,record[0],bnb[0],group_id)
                    else:
                        scorea = bnb[0][0]
                        scoreb = bnb[1][0]
                        if self.homology_algorithm != 'NB':
                            scorea = math.log(scorea)
                            scoreb = math.log(scoreb)
                        if abs(scorea - scoreb) >= diff:
                            #if self.homology_algorithm == 'DCMEGABLAST':
                            #    print '%s\t%s\t%s\t%s\t%s\t%s\t%s'%(record[0],bnb[0][1],bnb[0][2],bnb[0][0],bnb[1][1],bnb[1][2],bnb[1][0])
                            self.finish(job,record[0],bnb[0],group_id)
                job['rank'] = jr
        except StopIteration:
            pass
        job['rank'] = jr

class DbnRatioLabeller(HomologyRatioLabeller):
    prerequisites = [DbnClassifier]
    homology_algorithm = 'DCMEGABLAST'
    def __init__(self):
        self.fiot = BLASTLogIO

class BnRatioLabeller(HomologyRatioLabeller):
    prerequisites = [BnClassifier]
    homology_algorithm = 'BLASTN'
    def __init__(self):
        self.fiot = BLASTLogIO


class BxRatioLabeller(HomologyRatioLabeller):
    prerequisites = [BxClassifier]
    homology_algorithm = 'BLASTX'
    def __init__(self):
        self.fiot = BLASTLogIO

    
class NbRatioLabeller(HomologyRatioLabeller):
    prerequisites = [NbClassifier]
    homology_algorithm = 'NB'
    def __init__(self):
        self.fiot = NBLogIO

class UBxRatioLabeller(HomologyRatioLabeller):
    prerequisites = [UBxClassifier]
    homology_algorithm = 'UBLASTX'
    def __init__(self):
        self.fiot = BLASTLogIO

