"""
   Rapid identification of taxonomic assignments (RITA)
   This program executes and combines the results of several BLAST algorithms and
   the Naive Bayes algorithm.  Sequences that are classified early in the pipeline
   are not executed with later algorithms, saving runtime.
   
   This work has been submitted for publication:
   MacDonald Norman J, Parks Donovan H, and Beiko Robert G.  Rapid identification
   of taxonomic assignments. Submitted 2011.
   
   If you have any questions or bug reports, please let us know at <beiko@cs.dal.ca>.
   
   This program depends on the installation of the Fragment Classification Package
   (available at http://kiwi.cs.dal.ca/Software_main)

   job.py contains the datastructure that tracks the job specific parameters (where are the fragments,
   what is the definition of the pipeline for this execution).  There are many fields that are not used
   in the standalone version that are critical for the web site (e.g. session_id, job_uuid).
   
   Some of the methods here (e.g. add_runtime) seem strange, but again exist as they are functions that
   may interact with the database with the web version.  In the standalone version, most of the progress
   is kept in memory, whereas the web application version stores most of the information in the database.
"""

import uuid,random
from datetime import datetime,date,timedelta
import application
from taxonomy import Taxonomy

class Job(dict):
    """ This object tracks everything necessary to run the pipeline """
    def __init__(self,params={}):
        self.default_params =  {'job_id':None,
                                'session_id': None,
                                'job_uuid': None,
                                'job_name':'unnamed',
                                'status':'INITIALIZING',
                                'message':None,
                                'email':None,
                                'submit':None,
                                'start':None,
                                'end':None,
                                'last_updated':None,
                                'data_directory':'.',
                                'email_notified':None,
                                'blastn_evalue':1e-5,
                                'blastx_evalue':1e-5,
                                'ublastx_evalue':1e-5,
                                'dcmegablast_evalue':1e-5,
                                'dcmegablast_diff':20,
                                'blastn_diff':20,
                                'blastx_diff':20,
                                'ublastx_diff':20,
                                'nb_diff':20,
                                'rank':'PHYLUM',
                                'np':4,
                                'nb_processor_ratio':0.25,
                                'sequences_filepath':'sequences.fasta',
                                'scratch_directory':'.',
                                'store_sequences':1,
                                'n_mgsequences':None,
                                'n_ssusequences':None,
                                'pipeline_version':0.1,
                                'archived':0,
                                'sge_id':None,
                                'ip_address':None,
                                'nb_diff': 20,
                                'proxy':None
                                }
        # Initialize dynamic parameters

        self.default_params['job_uuid'] = uuid.uuid4() #random uuid to eliminate job guessing for other's jobs
        #self.default_params['job_id'] = random.randint(1000,9999)
        self.default_params['submit'] = datetime.now() + timedelta(hours=2)
        self.default_params['last_updated'] = datetime.now() + timedelta(hours=2)
        self.update(self.default_params.copy())
        self.update(params)
        self.output_buffer = []
        self.output_buffer_size = 1000
        self.fout = None
        self['taxonomy'] = Taxonomy(self['taxonomy_path'])
        self['finished_sequences'] = set()
        
    def get_pipeline_labellers(self):
        return self['pipeline']
    
    def get_finished(self):
        return self['finished_sequences']
    
    def add_finished(self,fragment_id):
        self['finished_sequences'].add(fragment_id)
    
    def write_classification(self,fragment_id,model,score,group,specific_model):
        if self.fout == None:
            self.fout = open(self['output_filename'],'w')
        
        rank = self['rank']
        ranks = ('SPECIES','GENUS','FAMILY','ORDER','CLASS','PHYLUM','DOMAIN')
        ranki = ranks.index(rank)
        #t = self['taxonomy']['specific_model']
        taxout = [self['taxonomy'].get(specific_model,x) for x in ranks[ranki:]]
        line = '%s\t%s\t%s\t%s\t%s\t%s\n'%(fragment_id,group+1,self['pipeline'][group],rank,'\t'.join(taxout),specific_model)
        #self.output_buffer.append('%s\t%s\t%s\t%s'%(fragment_id,model,score,group))
        self.fout.write(line)
        self.add_finished(fragment_id)
        #if len(self.output_buffer) > 10000:
        #    self.fout.write('%s\n'%('\n'.join(self.output_buffer)))
        #    self.output_buffer = []
        
        
            
    def add_runtime(self,name,seconds,nseqs):
        print '%s\t%ss'%(name,seconds)
        
    
    
    def __str__(self):
        s = []
        s.append("** %s **"%(self.get('job_name','unnamed job')))
        for p in sorted(self):
            s.append("%s:\t%s"%(p,self[p]))
        return '\n'.join(s)
                             
  
