"""
   Rapid identification of taxonomic assignments (RITA)
   This program executes and combines the results of several BLAST algorithms and
   the Naive Bayes algorithm.  Sequences that are classified early in the pipeline
   are not executed with later algorithms, saving runtime.
   
   This work has been submitted for publication:
   MacDonald Norman J, Parks Donovan H, and Beiko Robert G.  Rapid identification
   of taxonomic assignments. Submitted 2011.
   
   If you have any questions or bug reports, please let us know at <beiko@cs.dal.ca>.
   
   This program depends on the installation of the Fragment Classification Package
   (available at http://kiwi.cs.dal.ca/Software_main)

   trees.py contains objects for interfacing with MOTHUR (16s alignments), FastTree (16S trees)
   and an object for malipulating trees (TreeUtils)
   
   The TreeUtils object is dependent on a BioPython installation (Bio.Nexus.Trees) for parsing the 
   FastTree output.

"""

import os,sys
from taxonomy import Taxonomy
import application

class MothurIO(object):
    def get_alignment(self,job,input_filepath):
        
        fnoext = '.'.join(input_filepath.split('.')[:-1])
        command = '%s "#align.seqs(candidate=%s.fasta,template=%s,flip=True);'\
                      'merge.files(input=%s.align-%s,output=%s.all.align);'\
                      'filter.seqs(fasta=%s.all.align,hard=%s/Lane1349.gg.filter,vertical=True);"' % (job['mothur_path'],fnoext,job['mothur_16s_template'],fnoext,job['mothur_16s_alignment'],fnoext,fnoext,job['mothur_root'])
        os.system(command)
        
        
class FastTreeIO(object):
    def get_tree(self,job,input_filepath,output_filepath):
        a = application
        command = '%s -nt -gtr < %s > %s' % ( job['fasttree_path'],input_filepath,output_filepath)
        os.system(command)



class TreeUtils(object):
    
    def load_properties(self,properties_filename):
        return Taxonomy(properties_filename)
        #properties = {}
        #header = []
        #for ln, line in enumerate(open(properties_filename)):
        #    fields = line.strip().split('\t')
        #    if ln == 0:
        #        header = fields
        #    else:
        #        properties[fields[0]] = dict(zip(header,fields))
        #return properties
    
    def reroot_archaea_bacteria(self,properties_filename,input_filepath,output_filepath):
        from Bio.Nexus.Trees import Tree
        t = Tree(open(input_filepath).read())
        properties = self.load_properties(properties_filename)
        
        # Get both children of the root.Keep following child that is not
        # monophyletic until it is, and then reroot on that node.
        n = t.root
        finished = False
        group_not_to_root = None
        while not finished:
            children = t.node(n).succ
            sets = {}
            for c in children:
                sets[c] = set()
                for taxon in t.get_taxa(c):
                    if properties.get(taxon,'DOMAIN') != None:
                        sets[c].add(properties.get(taxon,'DOMAIN'))
            for c in children:
                if len(sets[c]) > 1:
                    n = c
                    for cb in children:
                        if len(sets[cb])== 1:
                            group_not_to_root = list(sets[cb])[0]
                    break
            else:
                finished = True
        found = False
        for c in children:
            for taxon in t.get_taxa(c):
                if properties.get(taxon,'DOMAIN') != None:
                    if properties.get(taxon,'DOMAIN') != group_not_to_root:
                        found = True
                        break
            if found:
                break
            
        outgroup = t.get_taxa(c)
        t.root_with_outgroup(outgroup)
        fout = open(output_filepath,"w")
        fout.write("%s"%(t.to_string(plain=False).split("=")[1].strip()))
        fout.close()

    def load_source_ids(self,mothur_16s_tree):
        return [line.split()[0].strip()[1:] for line in open(mothur_16s_tree) if line.startswith('>')]

    def get_consensus_index(self,properties,genomes):
        taxon = ('species','genus','family','order','class','phylum','domain')
        for cindex in range(len(taxon)):
            ga = genomes[0]
            for gb in genomes[1:]:
                if properties.get(ga,taxon[cindex].upper()) != properties.get(gb,taxon[cindex].upper()):
                    break # goes to next cindex in outer loop
            else:  # no breaks in inner loop, consenus found
                break #exit outer loop, consensus index found
        else: # no consensus, a mix of archaea and bacteria (this shouldn't be possible with a root between the two)
            raise Exception("Mixture of archaea and bacteria found.")
        return cindex
            
        

    def get_priors(self,properties_filename,model16salignment,tree_filename,query_idlist):
        #initialize priors of everyone to 0
        #each genome has a prior array of size 7, species to domain
        from Bio.Nexus.Trees import Tree
        tree = Tree(open(tree_filename).read())
        properties = self.load_properties(properties_filename)
        source_idlist = self.load_source_ids(model16salignment)
        priors = {}
        for source_id in source_idlist:
            priors[source_id] = [0,0,0,0,0,0,0]
        terminals = tree.get_terminals()
        tdict = {}
        source_idset = set(source_idlist)
        query_idset = set(query_idlist)
        for t in terminals:
            #print t
            if t == None:
                print 'ERROR'
                sys.exit()
            n = tree.node(t)
            tdict[n.data.taxon] = t
            #print n.data.taxon
            
        # find minimum subtree for each node
        # and add 1 to prior of each
        # Find consensus, and add prior to that
        for query_id in query_idlist:
            nid = tdict[query_id]
            source_children = set()
            while len(source_children) == 0:
                #print nid,query_id,tree.root
                nid = tree.node(nid).get_prev()
                children = set(tree.get_taxa(nid))
                source_children = children.intersection(source_idset)
                #print len(source_children)
                
                if len(source_children) > 0:
                    source_childrena = source_children
                    source_childrenb = set()
                    while len(source_childrenb - source_childrena) ==0:
                        nid = tree.node(nid).get_prev()
                        source_childrenb = set(tree.get_taxa(nid)).intersection(source_idset) # - query_idset
                        if tree.node(nid) == tree.root:
                            break
                    
                    source_children = source_childrenb
            cindex = self.get_consensus_index(properties,list(set(tree.get_taxa(nid)) - query_idset))
            for child in source_children:
                priors[child][cindex] += 1
        return priors
    
    
